import re
from typing import Any, Dict, List, Optional

import arrow
import pandas as pd
from investor8_sdk.models.standardized_financial import StandardizedFinancial
from pandas.core.frame import DataFrame
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from i8_terminal.common.formatting import format_number
from i8_terminal.common.layout import format_df
from i8_terminal.common.utils import similarity
from i8_terminal.config import APP_SETTINGS


def get_statements_codes() -> List[str]:
    return ["cash_flow_statement", "income_statement", "balance_sheet_statement"]


def get_statements_disp_name(stmt: str) -> Optional[str]:
    return {
        "cash_flow_statement": "Cash Flow Statement",
        "income_statement": "Income Statement",
        "balance_sheet_statement": "Balance Sheet Statement",
    }.get(stmt)


def get_period_types() -> List[str]:
    return ["FY", "Q", "TTM", "YTD"]


def get_period_type_disp_name(period: str) -> Optional[str]:
    return {
        "FY": "Annual (FY)",
        "Q": "Quarterly (Q)",
        "TTM": "TTM",
        "YTD": "YTD",
    }.get(period)


def find_similar_statement(statement: str) -> Optional[str]:
    default_indicators = {"inc": "income_statement", "bs": "balance_sheet_statement", "cf": "cash_flow_statement"}
    defualt_ind = default_indicators.get(statement.lower())
    if defualt_ind:
        return defualt_ind
    statement_dicts = {
        "cash": "cash_flow_statement",
        "cash_flow": "cash_flow_statement",
        "cash_flow_statement": "cash_flow_statement",
        "income": "income_statement",
        "income_statement": "income_statement",
        "balance": "balance_sheet_statement",
        "balance_sheet": "balance_sheet_statement",
        "balance_sheet_statement": "balance_sheet_statement",
    }
    best_match_similarity = 0.0
    best_match = ""
    for stmt in statement_dicts.keys():
        sim = similarity(statement.lower(), stmt.lower())
        if sim > best_match_similarity:
            best_match_similarity = sim
            best_match = stmt
    if best_match_similarity < APP_SETTINGS["metrics"]["similarity_threshold"]:
        return None

    return statement_dicts.get(best_match)


def parse_identifier(identifier: str, period_type: Optional[str]) -> Dict[str, str]:
    identifiers_dict = {}
    parsed_identifier = re.split("[-.,_]", identifier.replace(" ", ""))
    identifiers_dict["ticker"] = parsed_identifier[0]
    if len(parsed_identifier) == 3:
        # {ticker}_{fiscal_year}_{fiscal_period}
        identifiers_dict["fiscal_year"] = parsed_identifier[1]
        identifiers_dict["fiscal_period"] = (
            parsed_identifier[2] if parsed_identifier[2] not in ["Q4YTD", "Q4TTM"] else "FY"
        )
    elif len(parsed_identifier) == 2:
        # {ticker}_{fiscal_year}
        if parsed_identifier[1].isnumeric():
            identifiers_dict["fiscal_year"] = parsed_identifier[1]
            if not period_type or period_type.upper() == "FY":
                identifiers_dict["fiscal_period"] = "FY"
        # {ticker}_{fiscal_period}
        else:
            identifiers_dict["fiscal_period"] = parsed_identifier[1]

    return identifiers_dict


def render_metrics_table(table: Table, df: DataFrame, periods_list: List[str], col_width: int) -> Table:
    for p in periods_list:
        table.add_column(width=col_width, justify="right")
    for i, r in df.iterrows():
        if r["is_significant"]:
            table.add_row(f'[cyan]{r["name"]}[cyan]', *[f"[cyan]{d}[cyan]" for d in r[periods_list].values])
        else:
            table.add_row(r["name"], *r[periods_list].values)

    return table


def fin_df2Tree(df: DataFrame, header: Dict[str, List[str]], periods_list: List[str], title: str) -> Tree:
    col_width = 12
    tree = Tree(Panel(title, width=55))
    # Add header table to tree
    header_table = Table(
        width=85 + (col_width * (len(periods_list) - 1)), show_lines=False, show_header=False, box=None
    )
    header_table.add_column(width=64, style="magenta")
    for p in periods_list:
        header_table.add_column(width=col_width, justify="center", style="magenta")
    header_table.add_row("", *header["period"])
    header_table.add_row("Period End Date", *header["end_date"])
    tree.add(header_table)

    for sec_name, sec_metrics in df.sort_values(["section_order", "sub_section_order", "tag_order"]).groupby(
        "section_name", sort=False
    ):
        sec_branch = tree.add(f" {sec_name}")

        if sec_metrics["sub_section_order"].values[0]:  # If sub sections exists add new branch to tree
            for sub_sec_name, sub_sec_metrics in sec_metrics.sort_values(
                ["section_order", "sub_section_order", "tag_order"]
            ).groupby("sub_section_name", sort=False):
                sub_sec_branch = sec_branch.add(sub_sec_name)
                t = Table(
                    width=75 + (col_width * (len(periods_list) - 1)), show_lines=False, show_header=False, box=None
                )
                t.add_column(width=55)
                render_metrics_table(t, sub_sec_metrics, periods_list, col_width)
                sub_sec_branch.add(t)

        else:
            t = Table(width=79 + (col_width * (len(periods_list) - 1)), show_lines=False, show_header=False, box=None)
            t.add_column(width=59)
            render_metrics_table(t, sec_metrics, periods_list, col_width)
            sec_branch.add(t)

    return tree


def prepare_financials_df(
    financials_list: List[StandardizedFinancial],
    period_size: int,
    include_ticker: bool = False,
    exportize: Optional[bool] = False,
) -> Dict[str, Any]:
    # Get financial rows
    df = DataFrame([d.to_dict() for d in financials_list])[
        ["ticker", "fiscal_year", "fiscal_period", "type", "filing_date", "end_date", "financial_tags"]
    ]
    df = df.sort_values(by=["fiscal_year", "fiscal_period", "filing_date"], ascending=False)
    df = df.groupby(["ticker", "fiscal_year", "fiscal_period"]).head(1)

    fins_dict: Dict[str, Dict[str, Any]] = {}
    header_dict: Dict[str, List[str]] = {"end_date": [], "period": [], "ticker": []}
    for idx, fin in df[:period_size].iterrows():
        period = fin.fiscal_year
        period = f"{fin.ticker} {period}" if include_ticker else period
        header_dict["end_date"].append(arrow.get(fin.end_date).datetime.strftime("%d %b %Y"))
        fiscal_period = "Annual" if fin.fiscal_period == "FY" else fin.fiscal_period
        idf = f"{period}\n({fiscal_period})"  # Identifier: {ticker} {fiscal_year} \n{fiscal_period}
        header_dict["period"].append(idf)
        header_dict["ticker"].append(fin.ticker)
        fins_dict[idf] = {}
        for tag in fin.financial_tags:
            fins_dict[idf][tag["tag_name"]] = format_number(
                tag["value"], tag["unit"], humanize=True, exportize=exportize
            )

    df_rows = DataFrame(fins_dict.values(), index=fins_dict.keys()).T
    df_rows = df_rows.where(pd.notnull(df_rows), "-")  # Fill nan with `-`
    df_rows.index = df_rows.index.set_names(["tag"])
    df_rows = df_rows.reset_index()
    df_rows["tag"] = df_rows["tag"].apply(lambda x: x.split("_")[0])

    return {"header": header_dict, "data": df_rows}


def fin_df2export_df(df: pd.DataFrame, periods_list: List[str]) -> pd.DataFrame:
    col_names = {
        "name": "",
    }
    for column in periods_list:
        col_names[column] = column.replace("\n", " ")
    return format_df(df, col_names, {})


def available_fin_df2tree(df: pd.DataFrame, ticker: str) -> Tree:
    col_width = 25
    tree = Tree(Panel(f"{ticker} Financial Statements Coverage", width=40))
    statement_codes = get_statements_codes()
    statement_codes.sort(reverse=True)
    # Add header table to tree
    header_table = Table(
        width=50 + (col_width * (len(statement_codes) - 1)), show_lines=False, show_header=False, box=None
    )
    header_table.add_column(width=20, style="magenta")
    for p in statement_codes:
        header_table.add_column(width=col_width, justify="center", style="magenta")
    header_table.add_row(
        "Statement Code", *[get_statements_disp_name(statement_code) for statement_code in statement_codes]
    )
    tree.add(header_table)

    for fiscal_year, available_fins_year_df in df.sort_values(["fiscal_year"], ascending=False).groupby(
        "fiscal_year", sort=False
    ):
        year_branch = tree.add(f" {fiscal_year}")

        has_fy = False
        for period_type, period_type_df in available_fins_year_df.merge(
            DataFrame(get_period_types(), columns=["period_type"]), on="period_type", how="right"
        ).groupby("period_type", sort=False):
            period_type_df = period_type_df.merge(
                DataFrame(get_statements_codes(), columns=["statement_code"]), on="statement_code", how="right"
            )
            period_type_disp_name = f" {get_period_type_disp_name(period_type)}"
            t = Table(
                width=46 + (col_width * (len(statement_codes) - 1)),
                show_lines=False,
                show_header=False,
                box=None,
                style="cyan" if period_type == "FY" or period_type == "TTM" else None,
            )
            t.add_column(width=16)
            for st in statement_codes:
                t.add_column(width=col_width, justify="center")
            if period_type_df.empty:
                t.add_row(period_type_disp_name, *["❌", "❌", "❌"])
            else:
                row_text = []
                fiscal_period_df = period_type_df[["statement_code", "fiscal_period"]].groupby(
                    "statement_code", sort=False
                )
                if period_type == "FY":
                    for statement_code in get_statements_codes():
                        if pd.isnull(
                            fiscal_period_df.get_group(statement_code)["fiscal_period"].reset_index(drop=True)[0]
                        ):
                            row_text.append("❌")
                        else:
                            row_text.append("✔️")
                            has_fy = True
                else:
                    for statement_code in get_statements_codes():
                        if pd.isnull(
                            fiscal_period_df.get_group(statement_code)["fiscal_period"].reset_index(drop=True)[0]
                        ):
                            row_text.append(
                                "NA"
                                if (
                                    statement_code == "balance_sheet_statement"
                                    and (period_type == "TTM" or period_type == "YTD")
                                )
                                else "❌"
                            )
                        else:
                            available_fiscal_periods = []
                            for fiscal_period in fiscal_period_df.get_group(statement_code)["fiscal_period"]:
                                available_fiscal_periods.append(f"{fiscal_period[0:2]}")
                            available_fiscal_periods = list(dict.fromkeys(available_fiscal_periods))
                            available_fiscal_periods.sort()
                            row_text.append(
                                "Q1-Q4"
                                if available_fiscal_periods == ["Q1", "Q2", "Q3", "Q4"]
                                or (period_type == "TTM" and available_fiscal_periods == ["Q1", "Q2", "Q3"] and has_fy)
                                else ",".join(available_fiscal_periods)
                            )
                t.add_row(period_type_disp_name, *row_text)

            year_branch.add(t)

    return tree



from datetime import datetime
from typing import Any, Dict, List, Optional, cast

import click
import investor8_sdk
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from click.types import DateTime
from pandas import DataFrame
from plotly.subplots import make_subplots
from rich.console import Console

from i8_terminal.app.layout import get_plot_default_layout
from i8_terminal.app.plot_server import serve_plot
from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import get_click_command_path, pass_command
from i8_terminal.common.metrics import get_period_start_date
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import PlotType
from i8_terminal.types.chart_param_type import ChartParamType
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.period_type_param_type import PeriodTypeParamType
from i8_terminal.types.price_period_param_type import PricePeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def get_historical_metrics_df(
    tickers: List[str],
    metrics: List[str],
    period: str,
    period_type: str,
    from_date: Optional[str],
    to_date: Optional[str],
) -> DataFrame:
    metrics_period_type = (
        [f"{metric}.{period_type}" for metric in metrics] if period_type else [metric for metric in metrics]
    )
    if from_date:
        if not to_date:
            to_date = datetime.now().strftime("%Y-%m-%d")
        historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
            symbols=",".join(tickers),
            metrics=",".join(metrics_period_type),
            from_date=from_date,
            to_date=to_date,
        )
    else:
        from_date = get_period_start_date(period)
        to_date = datetime.now().strftime("%Y-%m-%d")
        historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
            symbols=",".join(tickers),
            metrics=",".join(metrics_period_type),
            from_date=from_date,
            to_date=to_date,
        )
    df = pd.DataFrame.from_records(
        [
            (ticker, metric, period_value.period, period_value.period_date_time, period_value.value)
            for ticker, metric_dict in historical_metrics.data.items()
            for metric, period_value_list in metric_dict.items()
            for period_value in period_value_list
        ],
        columns=["Ticker", "metric_name", "Period", "PeriodDateTime", "Value"],
    )
    metadata_df = pd.DataFrame([h.to_dict() for h in historical_metrics.metadata])
    df = pd.merge(df, metadata_df, on="metric_name")
    df.rename(columns={"display_name": "Metric"}, inplace=True)
    df["Value"] = df["Value"].astype(df["data_format"].values[0])
    return df


def create_fig(
    df: DataFrame,
    cmd_context: Dict[str, Any],
    tickers: List[str],
    chart_type: str,
) -> go.Figure:
    vertical_spacing = 0.02
    layout = dict(
        title=cmd_context["plot_title"],
        autosize=True,
        hovermode="closest",
        legend=dict(font=dict(size=11), orientation="v"),
        margin=dict(b=20, l=50, r=65),
    )
    metrics = list(set(df["Metric"]))
    rows_num = len(metrics)

    if rows_num == 2:
        row_width = [0.5, 0.5]
    else:
        row_width = [1]

    fig = make_subplots(
        rows=rows_num,
        cols=1,
        shared_xaxes=False if len(list(set(df["default_period_type"]))) > 1 else True,
        vertical_spacing=vertical_spacing,
        row_width=row_width,
    )

    for metric, metric_df in df.groupby("Metric"):
        metric_idx = metrics.index(metric)
        for ticker, ticker_df in metric_df.sort_values(["PeriodDateTime"]).groupby("Ticker"):
            idx = tickers.index(ticker)
            if chart_type == "bar":
                fig.add_trace(
                    go.Bar(
                        x=ticker_df["Period"],
                        y=ticker_df["Value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
            else:
                fig.add_trace(
                    go.Scatter(
                        x=ticker_df["Period"],
                        y=ticker_df["Value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
        if len(metrics) > 1:
            fig["layout"][f"xaxis{metric_idx+1}"]["dtick"] = round(len(metric_df["Period"]) / 10)

    fig.update_traces(hovertemplate="%{y} %{x}")
    fig.update_xaxes(
        rangeslider_visible=False,
        spikemode="across",
        spikesnap="cursor",
    )

    fig.update_layout(
        **layout,
        **get_plot_default_layout(),
        legend_title_text=None,
        xaxis_title=None,
        yaxis_title=None,
    )

    # Add yaxis titles
    for idx, r in enumerate(np.cumsum(row_width)[::-1]):
        fig["layout"][f"yaxis{idx+1}"]["title"] = metrics[idx]

    fig.update_annotations(
        dict(
            font_size=10,
            font_color="#525252",
        )
    )
    fig.update_yaxes(
        title_font_size=10,
    )

    return fig


@metrics.command()
@click.pass_context
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option(
    "--metrics",
    "-m",
    type=MetricParamType(),
    default="pe_ratio_ttm",
    help="Comma-separated list of daily metrics.",
)
@click.option(
    "--period",
    "-p",
    type=PricePeriodParamType(),
    default="1Y",
    help="Historical price period.",
)
@click.option(
    "--period_type",
    "-m",
    type=PeriodTypeParamType(),
    help="Period by which you want to view the report. Possible values are `FY` for yearly, `Q` for quarterly, and `TTM` for TTM reports.",
)
@click.option("--from_date", "-f", type=DateTime(), help="Histotical financials from date.")
@click.option("--to_date", "-t", type=DateTime(), help="Histotical financials to date.")
@click.option(
    "--chart_type",
    "-c",
    type=ChartParamType([("bar", "Bar chart"), ("line", "Line chart")]),
    default="line",
    help="Chart can be bar or line chart.",
)
@pass_command
def plot(
    ctx: click.Context,
    tickers: str,
    metrics: str,
    chart_type: str,
    period: str,
    period_type: str,
    from_date: Optional[datetime],
    to_date: Optional[datetime],
) -> None:
    """
    Compares and plots daily metrics of given companies. TICKERS is a comma-separated list of tickers.

    Examples:

    `i8 metrics plot --metrics price_to_earnings --period 5Y --tickers AMD,INTC,QCOM`
    """
    period = period.replace(" ", "").upper()
    metrics_list = metrics.replace(" ", "").split(",")
    if len(metrics_list) > 2:
        click.echo("You can enter up to 2 daily metrics.")
        return
    command_path_parsed_options_dict = {}
    if from_date:
        command_path_parsed_options_dict["--from_date"] = from_date.strftime("%Y-%m-%d")
    if to_date:
        command_path_parsed_options_dict["--to_date"] = to_date.strftime("%Y-%m-%d")
    command_path = get_click_command_path(ctx, command_path_parsed_options_dict)
    tickers_list = tickers.replace(" ", "").upper().split(",")
    if len(tickers_list) > 5:
        click.echo("You can enter up to 5 tickers.")
        return
    if not chart_type in ["bar", "line"]:
        click.echo(f"`{chart_type}` is not valid chart type.")
        return
    cmd_context = {
        "command_path": command_path,
        "tickers": tickers_list,
        "plot_type": PlotType.CHART.value,
    }

    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        df = get_historical_metrics_df(
            tickers_list, metrics_list, period, period_type, cast(str, from_date), cast(str, to_date)
        )
        for m in [*set(metric.split(".")[0] for metric in set(metrics.split(","))) - set(df["metric_name"])]:
            console.print(f"\nNo data found for metric {m} with selected tickers", style="yellow")
        cmd_context["plot_title"] = f"Historical {' and '.join(list(set(df['Metric'])))}"
        status.update("Generating plot...")
        fig = create_fig(df, cmd_context, tickers_list, chart_type)

    serve_plot(fig, cmd_context)



from typing import Any, Dict, List, Optional, cast

import click
import plotly.express as px
from click.types import DateTime
from pandas.core.frame import DataFrame
from plotly.graph_objects import Figure
from rich.console import Console

from i8_terminal.app.layout import get_date_range, get_plot_default_layout
from i8_terminal.app.plot_server import serve_plot
from i8_terminal.commands.price import price
from i8_terminal.common.cli import get_click_command_path, pass_command
from i8_terminal.common.price import (
    get_historical_price_df,
    get_historical_price_export_df,
)
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import PlotType, export_data, get_period_code
from i8_terminal.config import APP_SETTINGS
from i8_terminal.types.price_period_param_type import PricePeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def get_price_data(
    tickers: List[str], period_code: int, from_date: Optional[str], to_date: Optional[str]
) -> Optional[DataFrame]:
    hist_price_df = get_historical_price_df(tickers, period_code, from_date, to_date, pivot_value="change_perc")
    # FIXME: use realtime price in future
    # realtime_price_df = get_historical_price_df(tickers, 1)
    # df = hist_price_df.append(realtime_price_df)
    if hist_price_df is None:
        return None
    hist_price_df = hist_price_df.dropna()

    return hist_price_df.sort_values("Date")


def get_export_data(
    tickers: List[str], period_code: int, from_date: Optional[str], to_date: Optional[str]
) -> Optional[DataFrame]:
    compare_columns = {"change_perc": "Change (%)", "close": "Close"}
    export_price_df = get_historical_price_export_df(
        tickers, period_code, from_date, to_date, compare_columns=compare_columns
    )
    if export_price_df is None:
        return None
    return export_price_df.sort_values("Date")


def create_fig(df: DataFrame, period_code: int, cmd_context: Dict[str, Any], range_selector: bool = False) -> Figure:
    layout = dict(
        autosize=True,
        hovermode="closest",
        legend=dict(font=dict(size=11), orientation="h", yanchor="top", y=0.98, xanchor="left", x=0.01),
    )
    fig = px.line(
        df,
        x="Date",
        y=df.columns,
        title=cmd_context["plot_title"],
        labels={"value": "Change (%)"},
    )
    fig.update_xaxes(
        rangeslider_visible=False,
        spikemode="across",
        spikesnap="cursor",
    )
    if range_selector:
        fig.update_xaxes(rangeselector=get_date_range(period_code))
    fig.update_yaxes(tickformat=".0%")
    fig.update_layout(
        **layout,
        xaxis_title=None,
        **get_plot_default_layout(),
        margin=dict(b=15, l=90, r=20),
    )

    return fig


@price.command()
@click.pass_context
@click.option(
    "--period",
    "-p",
    type=PricePeriodParamType(),
    default="1M",
    help="Historical price period.",
)
@click.option("--from_date", "-f", type=DateTime(), help="Histotical price from date.")
@click.option("--to_date", "-t", type=DateTime(), help="Histotical price to date.")
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def compare(
    ctx: click.Context,
    period: str,
    from_date: Optional[DateTime],
    to_date: Optional[DateTime],
    tickers: str,
    export_path: Optional[str],
) -> None:
    """
    Compares historical prices of given companies. TICKERS is a comma-separated list of tickers.

    Examples:

    `i8 price compare --period 3Y --tickers AMD,INTC,QCOM`
    """
    command_path = get_click_command_path(ctx)
    period = period.replace(" ", "").upper()
    period_code = get_period_code(period)
    tickers_list = tickers.replace(" ", "").upper().split(",")
    plot_title = f"Comparison of {', '.join(tickers_list)} prices"
    plot_title = " and ".join(plot_title.rsplit(", ", 1))

    cmd_context = {
        "command_path": command_path,
        "tickers": tickers_list,
        "plot_title": plot_title,
        "plot_type": PlotType.CHART.value,
    }

    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        if export_path:
            df = get_export_data(tickers_list, period_code, cast(str, from_date), cast(str, to_date))
            if df is None:
                status.stop()
                console.print("No data found!")
                return
        else:
            df = get_price_data(tickers_list, period_code, cast(str, from_date), cast(str, to_date))
            if df is None:
                status.stop()
                console.print("No data found!")
                return
            status.update("Generating plot...")
            fig = create_fig(df, period_code, cmd_context)

    if export_path:
        export_data(
            df,
            export_path,
            index=True,
            column_width=20,
            column_format=APP_SETTINGS["styles"]["xlsx"]["price"]["column"],
        )
    else:
        serve_plot(fig, cmd_context)
